/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.bpmn2bpel;


import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.wsdl.Definition;
import javax.wsdl.WSDLException;
import javax.xml.bind.JAXBException;
import javax.xml.namespace.QName;
import org.ow2.orchestra.bpmn2bpel.components.SequenceComponent;
import org.ow2.orchestra.bpmn2bpel.exceptions.TransformerException;
import org.ow2.orchestra.bpmn2bpel.generators.BpelGenerator;
import org.ow2.orchestra.bpmn2bpel.transformers.EmptyComponentTransformer;
import org.ow2.orchestra.bpmn2bpel.transformers.FlowConditionLinkTransformer;
import org.ow2.orchestra.bpmn2bpel.transformers.FlowLinkTransformer;
import org.ow2.orchestra.bpmn2bpel.transformers.FlowTransformer;
import org.ow2.orchestra.bpmn2bpel.transformers.GatewayTransformer;
import org.ow2.orchestra.bpmn2bpel.transformers.OnEventTransformer;
import org.ow2.orchestra.bpmn2bpel.transformers.PickTransformer;
import org.ow2.orchestra.bpmn2bpel.transformers.RepeatTransformer;
import org.ow2.orchestra.bpmn2bpel.transformers.RepeatWhileTransformer;
import org.ow2.orchestra.bpmn2bpel.transformers.SequenceTransformer;
import org.ow2.orchestra.bpmn2bpel.transformers.SwitchTransformer;
import org.ow2.orchestra.bpmn2bpel.transformers.WhileTransformer;
import org.ow2.orchestra.jaxb.bpmn.Definitions;
import org.ow2.orchestra.jaxb.bpmn.TEndEvent;
import org.ow2.orchestra.jaxb.bpmn.TFlowNode;
import org.ow2.orchestra.jaxb.bpmn.TProcess;
import org.ow2.orchestra.jaxb.bpmn.TSequenceFlow;
import org.ow2.orchestra.jaxb.bpmn.TStartEvent;
import org.ow2.orchestra.util.Misc;
import org.xml.sax.SAXException;


/**
 * @author Guillaume Porcher
 */
public final class Transformer {
  private static final Logger LOG = Logger.getLogger(Transformer.class.getName());

  private final EmptyComponentTransformer emptyComponentTransformer = new EmptyComponentTransformer();
  private final FlowConditionLinkTransformer flowConditionLinkTransformer = new FlowConditionLinkTransformer();
  private final FlowLinkTransformer flowLinkTransformer = new FlowLinkTransformer();
  private final FlowTransformer flowTransformer = new FlowTransformer();
  private final GatewayTransformer gatewayTransformer = new GatewayTransformer();
  private final PickTransformer pickTransformer = new PickTransformer();
  private final RepeatTransformer repeatTransformer = new RepeatTransformer();
  private final RepeatWhileTransformer repeatWhileTransformer = new RepeatWhileTransformer();
  private final SequenceTransformer sequenceTransformer = new SequenceTransformer();
  private final SwitchTransformer switchTransformer = new SwitchTransformer();
  private final WhileTransformer whileTransformer = new WhileTransformer();
  private final OnEventTransformer onEventTransformer = new OnEventTransformer();


  public Map<String, byte[]> transform(final URL baseUrl) throws JAXBException, IOException, SAXException, WSDLException {
    return this.transform(baseUrl.openStream(), baseUrl);
  }

  /**
   * Parse a bpmn file and transform the model to bpel.
   *
   * @param bpmn bpmn is taken from this input stream
   * @param baseUrl baseUrl to resolve imports
   * @throws JAXBException
   * @throws IOException
   * @throws SAXException
   * @throws WSDLException
   */
  public Map<String, byte[]> transform(
          final InputStream bpmn,
          final URL baseUrl)
          throws JAXBException, IOException, SAXException, WSDLException {
    // parse bpmn definition
    final Definitions definitions = JaxbUtil.parseBPMNDefinitions(bpmn);


    // get process informations
    final BpmnMap bpmnMap = new BpmnMap();

    BpmnMap.getBpmnMap(definitions, bpmnMap, baseUrl);

    final Collection<TProcess> processes = bpmnMap.getProcesses().values();
    if (processes.size() != 1) {
      throw new TransformerException("BPMN file should contain exactly one process.");
    }
    final TProcess process = processes.iterator().next();
    final QName processQName = bpmnMap.getProcesses().keySet().iterator().next();
    // get start event from bpmn
    final Collection<TStartEvent> startEvents = bpmnMap.getStartEvents().values();
    if (startEvents.size() == 0) {
      throw new TransformerException("Processes with no startEvent are not supported.");
    } else if (startEvents.size() > 1) {
      throw new TransformerException("Processes with more than one startEvent are currently not supported.");
    }
    final TStartEvent startEvent = startEvents.iterator().next();

    // do transformation until we have the main sequence
    while (bpmnMap.getFlowNodes().size() != 1) {
      Misc.fastDynamicLog(Transformer.LOG, Level.FINEST, "Current status: %s", this.printSequenceFlows(bpmnMap, startEvent));
      if (!this.gatewayTransformer.replaceGateways(bpmnMap)
              && !this.emptyComponentTransformer.replaceEmptyComponents(bpmnMap)
              && !this.sequenceTransformer.replaceMaxSequence(bpmnMap)
              && !this.flowTransformer.replaceFlow(bpmnMap)
              && !this.switchTransformer.replaceSwitch(bpmnMap, processQName.getNamespaceURI())
              && !this.pickTransformer.replacePick(bpmnMap)
              && !this.whileTransformer.replaceWhile(bpmnMap)
              && !this.repeatTransformer.replaceRepeat(bpmnMap)
              && !this.repeatWhileTransformer.replaceRepeatWhile(bpmnMap)
              && !this.flowLinkTransformer.replaceFlow(bpmnMap)
              && !this.flowConditionLinkTransformer.replaceFlow(bpmnMap)
              && !this.onEventTransformer.replaceWithOnEvent(bpmnMap)) {
        break;
      }
    }
    final List<TFlowNode> sequenceNodes;
    if (bpmnMap.getFlowNodes().size() == 1 && bpmnMap.getFlowNodes().values().iterator().next() instanceof SequenceComponent) {
      sequenceNodes = ((SequenceComponent) bpmnMap.getFlowNodes().values().iterator().next()).getSequenceNodes();
    } else {
      sequenceNodes = null;
    }
    if (sequenceNodes == null
            || !(sequenceNodes.get(sequenceNodes.size() - 1) instanceof TEndEvent)
            || !(sequenceNodes.get(0) instanceof TStartEvent)) {
      throw new TransformerException(
              "Main sequence does not start with a startEvent and end with a endEvent, generation is currently not supported."
                      + Misc.LINE_SEPARATOR
                      + "Structure during last iteration was:"
                      + Misc.LINE_SEPARATOR
                      + this.printSequenceFlows(bpmnMap, startEvent, new ArrayList<TSequenceFlow>()));
    }

    final Map<String, Definition> partnerWSDLs = WSDLGenerator.generateWsdlDefinitions(bpmnMap, processQName);

    // Create BPEL process
    final org.ow2.orchestra.jaxb.bpel.Process p =
            new BpelGenerator().generateBpelProcess(bpmnMap, process, processQName, sequenceNodes, partnerWSDLs);

    // result files
    // generate wsdls
    final Map<String, byte[]> files = WSDLGenerator.writeWsdls(partnerWSDLs);


    // generate BPEL xml
    final ByteArrayOutputStream bpel = new ByteArrayOutputStream();
    JaxbUtil.writeBPEL(p, bpel);

    files.put("process.bpel", bpel.toByteArray());

    return files;
  }


  private String printSequenceFlows(final BpmnMap bpmnMap, final TFlowNode node) {
    final List<TSequenceFlow> seqFlows = new ArrayList<TSequenceFlow>();
    final String result = this.printSequenceFlows(bpmnMap, node, seqFlows);
    // DEBUG: if logs are enabled, check the right number of sequences were printed (and accessible from start event)
    if (seqFlows.size() != bpmnMap.getSequenceFlows().size()) {
      throw new IllegalStateException(
              "invalid number of sequenceFlows: expecting " + bpmnMap.getSequenceFlows().size() + " but was " + seqFlows.size());
    }
    return result;
  }

  private String printSequenceFlows(final BpmnMap bpmnMap, final TFlowNode node, final List<TSequenceFlow> alreadyDone) {
    String result = "";
    for (final TSequenceFlow tSeqFlow : bpmnMap.getOutgoingSequenceFlows(node)) {
      if (!alreadyDone.contains(tSeqFlow)) {
        alreadyDone.add(tSeqFlow);
        result += node.getId()
                + " (" + node + ") -> "
                + ((TFlowNode) tSeqFlow.getTargetRef()).getId() + " (" + tSeqFlow.getTargetRef() + ")"
                + Misc.LINE_SEPARATOR;
        result += this.printSequenceFlows(bpmnMap, (TFlowNode) tSeqFlow.getTargetRef(), alreadyDone);
      }
    }
    return result;
  }

//  private List<TFlowNode> getMainSequence(final BpmnMap bpmnMap, final TStartEvent startEvent) {
//    final List<TFlowNode> sequenceNodes = this.sequenceTransformer.getSequence(bpmnMap, startEvent);
//    if (!(sequenceNodes.get(sequenceNodes.size() - 1) instanceof TEndEvent)) {
//      return null;
//    }
//    return sequenceNodes;
//  }

}
