/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.bpmn2bpel.generators;

import java.io.IOException;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import javax.wsdl.Definition;
import javax.xml.XMLConstants;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.namespace.QName;
import org.ow2.orchestra.bpmn2bpel.BpmnMap;
import org.ow2.orchestra.bpmn2bpel.components.OnEventComponent;
import org.ow2.orchestra.bpmn2bpel.components.OnEventComponent.Type;
import org.ow2.orchestra.bpmn2bpel.exceptions.TransformerException;
import org.ow2.orchestra.jaxb.bpel.Assign;
import org.ow2.orchestra.jaxb.bpel.Condition;
import org.ow2.orchestra.jaxb.bpel.Copy;
import org.ow2.orchestra.jaxb.bpel.CorrelationSet;
import org.ow2.orchestra.jaxb.bpel.CorrelationSets;
import org.ow2.orchestra.jaxb.bpel.Elseif;
import org.ow2.orchestra.jaxb.bpel.EventHandlers;
import org.ow2.orchestra.jaxb.bpel.Flow;
import org.ow2.orchestra.jaxb.bpel.From;
import org.ow2.orchestra.jaxb.bpel.If;
import org.ow2.orchestra.jaxb.bpel.Invoke;
import org.ow2.orchestra.jaxb.bpel.OnEvent;
import org.ow2.orchestra.jaxb.bpel.OnMessage;
import org.ow2.orchestra.jaxb.bpel.PartnerLink;
import org.ow2.orchestra.jaxb.bpel.PartnerLinks;
import org.ow2.orchestra.jaxb.bpel.Pick;
import org.ow2.orchestra.jaxb.bpel.Receive;
import org.ow2.orchestra.jaxb.bpel.Scope;
import org.ow2.orchestra.jaxb.bpel.Sequence;
import org.ow2.orchestra.jaxb.bpel.TActivity;
import org.ow2.orchestra.jaxb.bpel.TActivityContainer;
import org.ow2.orchestra.jaxb.bpel.TCorrelation;
import org.ow2.orchestra.jaxb.bpel.TCorrelationWithPattern;
import org.ow2.orchestra.jaxb.bpel.TCorrelations;
import org.ow2.orchestra.jaxb.bpel.TCorrelationsWithPattern;
import org.ow2.orchestra.jaxb.bpel.TInitiate;
import org.ow2.orchestra.jaxb.bpel.TOnAlarmPick;
import org.ow2.orchestra.jaxb.bpel.TPattern;
import org.ow2.orchestra.jaxb.bpel.To;
import org.ow2.orchestra.jaxb.bpel.Variable;
import org.ow2.orchestra.jaxb.bpel.Variables;
import org.ow2.orchestra.jaxb.bpmn.TEndEvent;
import org.ow2.orchestra.jaxb.bpmn.TEventDefinition;
import org.ow2.orchestra.jaxb.bpmn.TExclusiveGateway;
import org.ow2.orchestra.jaxb.bpmn.TFlowNode;
import org.ow2.orchestra.jaxb.bpmn.TIntermediateCatchEvent;
import org.ow2.orchestra.jaxb.bpmn.TMessageEventDefinition;
import org.ow2.orchestra.jaxb.bpmn.TSequenceFlow;
import org.ow2.orchestra.jaxb.bpmn.TTimerEventDefinition;
import org.xml.sax.SAXException;


/**
 * @author Guillaume Porcher
 */
public class OnEventGenerator {
  public static final String ON_EVENT_PARTNER_LINK_TYPE = "internalOnEventPartnerLinkType";
  public static final String ON_EVENT_PROPERTY = "internalOnEventCorrelationProperty";
  public static final String ON_EVENT_PL = "internalOnEventPL";
  public static final String ON_EVENT_PARTNER_LINK_ROLE = "process";
  public static final String ON_EVENT_MSG_TYPE = "internalOnEventMsg";
  private static final String VAR_NAME = "inVar";
  private static final String CS_NAME = "internalEventCS";

  private final BpelGenerator bpelGenerator;

  private final Set<String> operations = new HashSet<String>();


  OnEventGenerator() {
    this.bpelGenerator = new BpelGenerator();
  }

  /**
   * @param bpelGenerator
   */
  OnEventGenerator(final BpelGenerator bpelGenerator) {
    this.bpelGenerator = bpelGenerator;
  }

  /**
   * @param bpmnMap
   * @param onEventComponent
   * @param processQName
   * @return
   */
  TActivity createOnEventActivity(
          final BpmnMap bpmnMap,
          final OnEventComponent onEventComponent,
          final QName processQName,
          final Map<String, Definition> partnerWsdls) throws JAXBException, IOException, SAXException {
    // create bpel scope
    final Scope scope = new Scope();
    // add partnerlink
    PartnerLinks partnerLinks = new PartnerLinks();
    scope.setPartnerLinks(partnerLinks);
    PartnerLink partnerLink = new PartnerLink();
    partnerLink.setName(ON_EVENT_PL);
    partnerLink.setMyRole(ON_EVENT_PARTNER_LINK_ROLE);
    partnerLink.setPartnerRole(ON_EVENT_PARTNER_LINK_ROLE);

    partnerLink.setPartnerLinkType(new QName(processQName.getNamespaceURI(), ON_EVENT_PARTNER_LINK_TYPE, processQName.getPrefix()));
    partnerLinks.getPartnerLinks().add(partnerLink);
    // add variable
    Variables variables = new Variables();
    scope.setVariables(variables);
    Variable variable = new Variable();
    variable.setName(VAR_NAME);
    variable.setMessageType(new QName(processQName.getNamespaceURI(), ON_EVENT_MSG_TYPE, processQName.getPrefix()));
    variables.getVariables().add(variable);
    // Add correlationSet
    CorrelationSets correlationSets = new CorrelationSets();
    scope.setCorrelationSets(correlationSets);
    CorrelationSet correlationSet = new CorrelationSet();
    correlationSet.setName(CS_NAME);
    correlationSet.getProperties().add(new QName(processQName.getNamespaceURI(), ON_EVENT_PROPERTY ,processQName.getPrefix()));
    correlationSets.getCorrelationSets().add(correlationSet);

    EventHandlers eventHandlers = new EventHandlers();
    scope.setEventHandlers(eventHandlers);

    // create OnEvents
    Set<OnEventComponent> onEventComponents = new HashSet<OnEventComponent>();
    getOnEventComponents(onEventComponent, onEventComponents);
    for (OnEventComponent comp : onEventComponents) {
      OnEvent onEvent = new OnEvent();
      onEvent.setPartnerLink(ON_EVENT_PL);
      onEvent.setVariable(VAR_NAME);
      onEvent.setMessageType(new QName(processQName.getNamespaceURI(), ON_EVENT_MSG_TYPE, processQName.getPrefix()));
      onEvent.setOperation(comp.getId());
      onEvent.setCorrelations(createCorrelations());

      eventHandlers.getOnEvents().add(onEvent);
      Scope onEventScope = new Scope();
      onEvent.setScope(onEventScope);
      if (comp.getType().equals(Type.PARALLEL_FORK)) {
        // parallel split => create a flow
        Flow flow = new Flow();
        onEventScope.setFlow(flow);
        for (TSequenceFlow outSeqFlow : comp.getOutSequenceFlows()) {
          flow.getAssignsAndCompensatesAndCompensateScopes().add(getNextActivity(onEventComponent, comp, outSeqFlow));
        }
      } else if (comp.getType().equals(Type.PARALLEL_JOIN)) {
        // parallel join
        onEvent.setOperation(comp.getInSequenceFlows().get(0).getId() + "_" + comp.getId());
        Sequence seq = new Sequence();
        onEventScope.setSequence(seq);
        Flow flow = new Flow();
        seq.getAssignsAndCompensatesAndCompensateScopes().add(flow);
        for (int i = 1; i < comp.getInSequenceFlows().size(); i++) {
          TSequenceFlow seqFlow = comp.getInSequenceFlows().get(i);
          Receive r = new Receive();
          r.setPartnerLink(ON_EVENT_PL);
          r.setVariable(VAR_NAME);
          r.setCorrelations(createCorrelations());
          r.setOperation(seqFlow.getId() + "_" + comp.getId());
          this.operations.add(r.getOperation());

          flow.getAssignsAndCompensatesAndCompensateScopes().add(r);
        }
        TSequenceFlow outSeqFlow = comp.getOutSequenceFlows().get(0);
        seq.getAssignsAndCompensatesAndCompensateScopes().add(getNextActivity(onEventComponent, comp, outSeqFlow));
      } else if (comp.getType().equals(Type.EXCLUSIVE_FORK)) {
        // create bpel if
        final If ifAct = new If();
        onEventScope.setIf(ifAct);
        boolean isFirstConditionSet = false;

        // fill the sequence
        for (TSequenceFlow sequenceFlow : comp.getOutSequenceFlows()) {
          final Invoke invoke = this.getNextActivity(onEventComponent, comp, sequenceFlow);
          if (!sequenceFlow.equals(((TExclusiveGateway)comp.getOnEventNode()).getDefault())) {
            if (sequenceFlow.getConditionExpression() == null) {
              throw new TransformerException("Outgoing sequence flow of an exclusive gateway should have a condition");
            }
            final Condition condition = new Condition();
            condition.getContent().addAll(
                    this.bpelGenerator.getXPathExpression(sequenceFlow.getConditionExpression().getContent()));
            if (!isFirstConditionSet) {
              ifAct.setInvoke(invoke);
              ifAct.setCondition(condition);
              isFirstConditionSet = true;
            } else {
              final Elseif elseif = new Elseif();
              elseif.setInvoke(invoke);
              elseif.setCondition(condition);
              ifAct.getElseives().add(elseif);
            }
          } else {
            final TActivityContainer els = new TActivityContainer();
            els.setInvoke(invoke);
            ifAct.setElse(els);
          }
        }
      } else if (comp.getType().equals(Type.EXCLUSIVE_JOIN)) {
        TSequenceFlow outSeqFlow = comp.getOutSequenceFlows().get(0);
        onEventScope.setInvoke(getNextActivity(onEventComponent, comp, outSeqFlow));
      } else if (comp.getType().equals(Type.ACTIVITY)) {
        Sequence seq = new Sequence();
        onEventScope.setSequence(seq);
        seq.getAssignsAndCompensatesAndCompensateScopes().add(this.bpelGenerator.createBpelActivity(comp.getOnEventNode(), bpmnMap, processQName, partnerWsdls));
        TSequenceFlow outSeqFlow = comp.getOutSequenceFlows().get(0);
        seq.getAssignsAndCompensatesAndCompensateScopes().add(getNextActivity(onEventComponent, comp, outSeqFlow));
      } else if (comp.getType().equals(Type.PICK)) {
        // create bpel pick
        final Pick pick = new Pick();
        PickGenerator pickGenerator = new PickGenerator(this.bpelGenerator);
        final Variables onEventVariables = new Variables();
        onEventScope.setVariables(onEventVariables);

        // fill the events
        for (int i = 0; i < comp.getPickOutNodes().size(); i ++) {
          final Invoke activity = getNextActivity(onEventComponent, comp, comp.getOutSequenceFlows().get(i));
          final TIntermediateCatchEvent eventNode = (TIntermediateCatchEvent) comp.getPickOutNodes().get(i);
          for (final JAXBElement<? extends TEventDefinition> eventDefinition : eventNode.getEventDefinitions()) {
            if (eventDefinition.getValue() instanceof TMessageEventDefinition) {
              // create on message
              final OnMessage onMessage =
                      pickGenerator.createPickOnMessage(
                              bpmnMap,
                              processQName,
                              activity,
                              (TMessageEventDefinition) eventDefinition.getValue(),
                              eventNode,
                              onEventScope);
              pick.getOnMessages().add(onMessage);
            } else if (eventDefinition.getValue() instanceof TTimerEventDefinition) {
              final TOnAlarmPick onAlarm =
                      pickGenerator.createPickOnAlarm(activity, (TTimerEventDefinition) eventDefinition.getValue());
              pick.getOnAlarms().add(onAlarm);
            }
          }
        }
        onEventScope.setPick(pick);
        if (onEventVariables.getVariables().isEmpty()) {
          onEventScope.setVariables(null);
        }
      }
      this.operations.add(onEvent.getOperation());
    }
    // create main scope sequence
    final Sequence sequence = new Sequence();
    scope.setSequence(sequence);
    // initialize variable
    final Assign assign = new Assign();
    Copy copy = new Copy();
    From from = new From();
    from.getOtherAttributes().put(new QName(XMLConstants.XMLNS_ATTRIBUTE_NS_URI, "orchestra", XMLConstants.XMLNS_ATTRIBUTE), "http://orchestra.ow2.org");
    from.getContent().add("orchestra:instance-uuid()");
    To to = new To();
    to.getContent().add("$" + VAR_NAME + ".payload");
    copy.setFrom(from);
    copy.setTo(to);
    assign.getCopiesAndExtensionAssignOperations().add(copy);
    sequence.getAssignsAndCompensatesAndCompensateScopes().add(assign);
    // initial onEvent invoke
    final Invoke invoke = new Invoke();
    invoke.setPartnerLink(ON_EVENT_PL);
    invoke.setOperation(onEventComponent.getId());
    invoke.setInputVariable(VAR_NAME);
    invoke.setCorrelations(createCorrelationsWithPattern());
    invoke.getCorrelations().getCorrelations().get(0).setInitiate(TInitiate.YES);
    sequence.getAssignsAndCompensatesAndCompensateScopes().add(invoke);
    // final receive
    final Receive receive = new Receive();
    receive.setOperation("end_" + onEventComponent.getId());
    receive.setPartnerLink(ON_EVENT_PL);
    receive.setVariable(VAR_NAME);
    receive.setCorrelations(createCorrelations());
    this.operations.add(receive.getOperation());

    sequence.getAssignsAndCompensatesAndCompensateScopes().add(receive);
    return scope;
  }

  private TCorrelations createCorrelations() {
    TCorrelations tCorrelations = new TCorrelations();
    TCorrelation tCorrelation = new TCorrelation();
    tCorrelation.setSet(CS_NAME);
    tCorrelation.setInitiate(TInitiate.NO);
    tCorrelations.getCorrelations().add(tCorrelation);
    return tCorrelations;
  }

  private TCorrelationsWithPattern createCorrelationsWithPattern() {
    TCorrelationsWithPattern tCorrelations = new TCorrelationsWithPattern();
    TCorrelationWithPattern tCorrelation = new TCorrelationWithPattern();
    tCorrelation.setSet(CS_NAME);
    tCorrelation.setInitiate(TInitiate.NO);
    tCorrelation.setPattern(TPattern.REQUEST);
    tCorrelations.getCorrelations().add(tCorrelation);
    return tCorrelations;
  }

  private Invoke getNextActivity(final OnEventComponent mainOnEventComponent, final OnEventComponent currentOnEventComponent, final TSequenceFlow outSeqFlow) {
    if (outSeqFlow.getTargetRef() instanceof OnEventComponent) {
      OnEventComponent next = (OnEventComponent) outSeqFlow.getTargetRef();
      Invoke i = createInvokeForOnEvent(next, currentOnEventComponent);
      return i;
    } else if (outSeqFlow.getTargetRef() instanceof TEndEvent) {
      Invoke i = new Invoke();
      i.setPartnerLink(ON_EVENT_PL);
      i.setInputVariable(VAR_NAME);
      i.setOperation("end_" + mainOnEventComponent.getId());
      return i;
    } else {
      throw new TransformerException("Invalid type: " + outSeqFlow.getTargetRef().getClass().getName());
    }
  }

  private void getOnEventComponents(OnEventComponent onEventComponent, Set<OnEventComponent> result) {
    result.add(onEventComponent);
    for (TSequenceFlow outSequenceFlow : onEventComponent.getOutSequenceFlows()) {
      if ((outSequenceFlow.getTargetRef() instanceof OnEventComponent) && !result.contains(outSequenceFlow.getTargetRef())) {
        getOnEventComponents((OnEventComponent) outSequenceFlow.getTargetRef(), result);
      }
    }
  }

  private Invoke createInvokeForOnEvent(OnEventComponent eventToInvoke, TFlowNode invokerActivity) {
    final Invoke invoke = new Invoke();
    invoke.setPartnerLink(ON_EVENT_PL);
    invoke.setInputVariable(VAR_NAME);
    if (eventToInvoke.getType().equals(Type.PARALLEL_JOIN)) {
      for (TSequenceFlow seqFlow : eventToInvoke.getInSequenceFlows()) {
        if (invokerActivity.equals(seqFlow.getSourceRef())) {
          invoke.setOperation(seqFlow.getId() + "_" + eventToInvoke.getId());
        }
      }
    } else {
      invoke.setOperation(eventToInvoke.getId());
    }
    return invoke;
  }

  public Set<String> getOperations() {
    return operations;
  }
}
