/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.bpmn2bpel.generators;

import java.io.IOException;
import java.util.List;
import java.util.Map;

import javax.wsdl.Definition;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.namespace.QName;

import org.ow2.orchestra.bpmn2bpel.BpelUtil;
import org.ow2.orchestra.bpmn2bpel.BpmnMap;
import org.ow2.orchestra.bpmn2bpel.components.PickComponent;
import org.ow2.orchestra.bpmn2bpel.exceptions.TransformerException;
import org.ow2.orchestra.jaxb.bpel.Assign;
import org.ow2.orchestra.jaxb.bpel.FromParts;
import org.ow2.orchestra.jaxb.bpel.OnMessage;
import org.ow2.orchestra.jaxb.bpel.Pick;
import org.ow2.orchestra.jaxb.bpel.Scope;
import org.ow2.orchestra.jaxb.bpel.Sequence;
import org.ow2.orchestra.jaxb.bpel.TActivity;
import org.ow2.orchestra.jaxb.bpel.TDurationExpr;
import org.ow2.orchestra.jaxb.bpel.TOnAlarmPick;
import org.ow2.orchestra.jaxb.bpel.Until;
import org.ow2.orchestra.jaxb.bpel.Variable;
import org.ow2.orchestra.jaxb.bpel.Variables;
import org.ow2.orchestra.jaxb.bpmn.DataOutputAssociation;
import org.ow2.orchestra.jaxb.bpmn.TEventDefinition;
import org.ow2.orchestra.jaxb.bpmn.TFlowNode;
import org.ow2.orchestra.jaxb.bpmn.TIntermediateCatchEvent;
import org.ow2.orchestra.jaxb.bpmn.TMessageEventDefinition;
import org.ow2.orchestra.jaxb.bpmn.TTimerEventDefinition;
import org.xml.sax.SAXException;


/**
 * @author Guillaume Porcher
 */
class PickGenerator {

  private final BpelGenerator bpelGenerator;

  PickGenerator() {
    this.bpelGenerator = new BpelGenerator();
  }

  /**
   * @param bpelGenerator
   */
  PickGenerator(final BpelGenerator bpelGenerator) {
    this.bpelGenerator = bpelGenerator;
  }

  TActivity createPickActivity(
      final BpmnMap bpmnMap,
      final PickComponent pickComponent,
      final QName processQName,
      final Map<String, Definition> partnerWsdls) throws JAXBException, IOException, SAXException {
    // create bpel pick
    final Pick pick = new Pick();
    // create scope to store onMessage anonymous variables
    final Scope scope = new Scope();
    final Variables variables = new Variables();
    scope.setVariables(variables);

    final List<TFlowNode> pickNodes = pickComponent.getNodes();
    // fill the events
    for (int i = 0; i < pickNodes.size(); i += 2) {
      final TActivity activity = this.bpelGenerator.createBpelActivity(pickNodes.get(i + 1), bpmnMap, processQName, partnerWsdls);
      final TIntermediateCatchEvent eventNode = (TIntermediateCatchEvent) pickNodes.get(i);
      for (final JAXBElement<? extends TEventDefinition> eventDefinition : eventNode.getEventDefinitions()) {
        if (eventDefinition.getValue() instanceof TMessageEventDefinition) {
          // create on message
          final OnMessage onMessage =
            this.createPickOnMessage(
                bpmnMap,
                processQName,
                activity,
                (TMessageEventDefinition) eventDefinition.getValue(),
                eventNode,
                scope);
          pick.getOnMessages().add(onMessage);
        } else if (eventDefinition.getValue() instanceof TTimerEventDefinition) {
          final TOnAlarmPick onAlarm =
            this.createPickOnAlarm(activity, (TTimerEventDefinition) eventDefinition.getValue());
          pick.getOnAlarms().add(onAlarm);
        }
      }
    }
    if (scope.getVariables().getVariables().isEmpty()) {
      return pick;
    } else {
      scope.setPick(pick);
      return scope;
    }
  }

  TOnAlarmPick createPickOnAlarm(final TActivity activity, final TTimerEventDefinition timerDef) {
    // create onAlarm
    final TOnAlarmPick onAlarm = new TOnAlarmPick();
    if (timerDef.getTimeDuration() != null) {
      final TDurationExpr duration = new TDurationExpr();
      duration.getContent().addAll(this.bpelGenerator.getXPathExpression(timerDef.getTimeDuration().getContent()));
      onAlarm.setFor(duration);
    } else if (timerDef.getTimeDate() != null) {
      final Until until = new Until();
      until.getContent().addAll(this.bpelGenerator.getXPathExpression(timerDef.getTimeDate().getContent()));
      onAlarm.setUntil(until);
    }
    BpelUtil.setChildActivity(onAlarm, activity);
    return onAlarm;
  }

  OnMessage createPickOnMessage(
          final BpmnMap bpmnMap,
          final QName processQName,
          final TActivity activity,
          final TMessageEventDefinition messageEventDefinition,
          final TIntermediateCatchEvent eventNode,
          final Scope scope) {
    if (messageEventDefinition.getOperationRef() == null) {
      throw new TransformerException("message events must define an operation.");
    }
    final QName operationQName = messageEventDefinition.getOperationRef();

    final OperationReference receiveOperation =
      this.bpelGenerator.getProcessOperationReference(bpmnMap, processQName, operationQName);
    if (receiveOperation == null) {
      throw new TransformerException("operation " + operationQName + " not found");
    }
    final OnMessage onMessage = new OnMessage();
    onMessage.setOperation(receiveOperation.getOperation().getName());
    onMessage.setPartnerLink(receiveOperation.getPartnerLinkName());
    onMessage.setPortType(receiveOperation.getPortTypeQName());

    final List<DataOutputAssociation> dataOutputAssociations = eventNode.getDataOutputAssociations();
    if (!dataOutputAssociations.isEmpty()) {
      final Assign assign = new Assign();
      final Object dataObject = this.bpelGenerator.getOutputData(dataOutputAssociations, receiveOperation, assign, true, bpmnMap);
      if (dataObject instanceof FromParts) {
        onMessage.setFromParts((FromParts) dataObject);
      } else if (dataObject instanceof Variable) {
        final Variable outVariable = (Variable) dataObject;
        scope.getVariables().getVariables().add(outVariable);
        onMessage.setVariable(outVariable.getName());
      }
      if (assign.getCopiesAndExtensionAssignOperations().isEmpty()) {
        BpelUtil.setChildActivity(onMessage, activity);
      } else {
        final Sequence sequence = new Sequence();
        sequence.getAssignsAndCompensatesAndCompensateScopes().add(assign);
        sequence.getAssignsAndCompensatesAndCompensateScopes().add(activity);
        onMessage.setSequence(sequence);
      }
    } else {
      BpelUtil.setChildActivity(onMessage, activity);
    }
    return onMessage;
  }

}
