/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.bpmn2bpel.transformers;

import java.util.List;
import java.util.ListIterator;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.ow2.orchestra.bpmn2bpel.BpmnMap;
import org.ow2.orchestra.bpmn2bpel.components.EmptyComponent;
import org.ow2.orchestra.bpmn2bpel.components.FlowComponent;
import org.ow2.orchestra.bpmn2bpel.components.JoinConditionComponent;
import org.ow2.orchestra.bpmn2bpel.components.LinkComponent;
import org.ow2.orchestra.jaxb.bpmn.TFlowNode;
import org.ow2.orchestra.jaxb.bpmn.TGateway;
import org.ow2.orchestra.util.Misc;


/**
 * @author Guillaume Porcher
 */
public class EmptyComponentTransformer {

  private static final Logger LOG = Logger.getLogger(EmptyComponentTransformer.class.getName());

  public boolean replaceEmptyComponents(final BpmnMap bpmnMap) {
    Misc.fastDynamicLog(EmptyComponentTransformer.LOG, Level.INFO, "Trying to replace empty components");
    for (final TFlowNode flowNode : bpmnMap.getFlowNodes().values()) {
      if (flowNode instanceof EmptyComponent) {
        // FlowNode is an empty component
        final TFlowNode previousNode = (TFlowNode) bpmnMap.getIncomingSequenceFlows(flowNode).get(0).getSourceRef();
        final TFlowNode nextNode = (TFlowNode) bpmnMap.getOutgoingSequenceFlows(flowNode).get(0).getTargetRef();
        final List<LinkComponent> outgoingLinkComponents = bpmnMap.getOutgoingLinks(flowNode);
        final List<LinkComponent> incomingLinkComponents = bpmnMap.getIncomingLinks(flowNode);

        boolean remove = false;
        if (!(previousNode instanceof TGateway) && incomingLinkComponents.isEmpty()) {
          // merge empty with previous node
          // redirect links
          for (final LinkComponent linkComponent : outgoingLinkComponents) {
            linkComponent.setSourceNode(previousNode);
          }
          // remove node
          remove = true;
        }
        if (!(nextNode instanceof TGateway) && outgoingLinkComponents.isEmpty()) {
          // merge empty with next node
          // redirect links
          for (final LinkComponent linkComponent : incomingLinkComponents) {
            linkComponent.setTargetNode(nextNode);
          }
          // update nextNode joinCondition
          JoinConditionComponent.addNewAndCondition(bpmnMap, bpmnMap.getJoinCondition(flowNode), nextNode);
          // remove node
          remove = true;
        }

        if (remove) {
          // remove node
          bpmnMap.changeSequenceFlowTarget(bpmnMap.getIncomingSequenceFlows(flowNode).get(0), nextNode);
          bpmnMap.removeNode(flowNode);
          return true;
        }
      } else if (flowNode instanceof FlowComponent) {
        final FlowComponent flowComponent = (FlowComponent) flowNode;
        for (final TFlowNode childNode : flowComponent.getFlowNodes()) {
          if (childNode instanceof EmptyComponent) {
            final List<LinkComponent> outgoingLinkComponents = bpmnMap.getOutgoingLinks(childNode);
            final List<LinkComponent> incomingLinkComponents = bpmnMap.getIncomingLinks(childNode);


            if ((incomingLinkComponents.isEmpty() && !outgoingLinkComponents.isEmpty())
                    || (!incomingLinkComponents.isEmpty() && outgoingLinkComponents.isEmpty())) {
              // only one link to empty, empty can be removed with all links.
              for (final LinkComponent linkComponent : incomingLinkComponents) {
                if (flowComponent.getFlowNodes().contains(linkComponent.getSourceNode())) {
                  bpmnMap.getLinks().remove(linkComponent);
                  JoinConditionComponent.removeLink(bpmnMap, linkComponent);
                }
              }
              // only one link from empty, empty can be removed with all links.
              for (final LinkComponent linkComponent : outgoingLinkComponents) {
                if (flowComponent.getFlowNodes().contains(linkComponent.getTargetNode())) {
                  bpmnMap.getLinks().remove(linkComponent);
                  JoinConditionComponent.removeLink(bpmnMap, linkComponent);
                }
              }
            } else {
              for (final LinkComponent linkComponent : bpmnMap.getOutgoingLinks(childNode)) {
                if (flowComponent.getFlowNodes().contains(linkComponent.getTargetNode())) {
                  // linkComponent can be merged
                  for (final LinkComponent incomingLink : bpmnMap.getIncomingLinks(childNode)) {
                    incomingLink.setTargetNode(linkComponent.getTargetNode());
                    if (linkComponent.getTransitionCondition() != null) {
                      if (incomingLink.getTransitionCondition() == null) {
                        incomingLink.setTransitionCondition(linkComponent.getTransitionCondition());
                      } else {
                        final String condition =
                                "(" + this.getCondition(linkComponent) + ") and (" + this.getCondition(incomingLink) + ")";
                        incomingLink.getTransitionCondition().getContent().clear();
                        incomingLink.getTransitionCondition().getContent().add(condition);
                      }
                    }
                  }
                  // update target node joinCondition
                  JoinConditionComponent.replaceCondition(
                          bpmnMap,
                          linkComponent,
                          bpmnMap.getJoinCondition(childNode),
                          linkComponent.getTargetNode());
                  bpmnMap.getLinks().remove(linkComponent);
                  JoinConditionComponent.removeLink(bpmnMap, linkComponent);
                }
              }
            }
            // remove empty in a flow if there is no links
            if (bpmnMap.getOutgoingLinks(childNode).isEmpty() && bpmnMap.getIncomingLinks(childNode).isEmpty()) {
              flowComponent.getFlowNodes().remove(childNode);
              return true;
            }

          }
        }
      }
    }

    return false;
  }

  private String getCondition(final LinkComponent linkComponent) {
    final ListIterator<Object> iterator = linkComponent.getTransitionCondition().getContent().listIterator();
    while (iterator.hasNext()) {
      final Object o = iterator.next();
      if (o instanceof String) {
        return (String) o;
      }
    }
    return "true";
  }
}
