/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.bpmn2bpel.transformers;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.namespace.QName;
import org.ow2.orchestra.bpmn2bpel.BpmnMap;
import org.ow2.orchestra.bpmn2bpel.TFlowNodeComparator;
import org.ow2.orchestra.bpmn2bpel.components.SwitchComponent;
import org.ow2.orchestra.jaxb.bpmn.TExclusiveGateway;
import org.ow2.orchestra.jaxb.bpmn.TExpression;
import org.ow2.orchestra.jaxb.bpmn.TFlowNode;
import org.ow2.orchestra.jaxb.bpmn.TGateway;
import org.ow2.orchestra.jaxb.bpmn.TSequenceFlow;
import org.ow2.orchestra.util.Misc;


/**
 * @author Guillaume Porcher
 */
public final class SwitchTransformer {

  private static final Logger LOG = Logger.getLogger(SwitchTransformer.class.getName());

  /**
   * Get and replace a switch (if/else)
   *
   * @param bpmnMap
   * @return
   */
  public boolean replaceSwitch(final BpmnMap bpmnMap, final String processNamespace) {
    Misc.fastDynamicLog(SwitchTransformer.LOG, Level.INFO, "Trying to replace switch");
    List<TSequenceFlow> sequenceFlows = Collections.emptyList();
    for (final TFlowNode flowNode : bpmnMap.getFlowNodes().values()) {
      if (flowNode instanceof TExclusiveGateway) {
        sequenceFlows = this.getSwitch(bpmnMap, flowNode);
        if (!sequenceFlows.isEmpty()) {
          break;
        }
      }
    }
    if (sequenceFlows.isEmpty() || sequenceFlows.size() < 2) {
      return false;
    }
    final TSequenceFlow firstSequenceFlow = sequenceFlows.get(0);

    final TExclusiveGateway startGateway = (TExclusiveGateway) firstSequenceFlow.getSourceRef();
    final TFlowNode firstNode = (TFlowNode) firstSequenceFlow.getTargetRef();
    final TExclusiveGateway endGateway =
            (TExclusiveGateway) bpmnMap.getOutgoingSequenceFlows(firstNode).iterator().next().getTargetRef();

    // get default sequence flow
    TSequenceFlow defaultSequenceFlow = null;
    if (startGateway.getDefault() != null && sequenceFlows.contains(startGateway.getDefault())) {
      defaultSequenceFlow = (TSequenceFlow) startGateway.getDefault();
    }
    // flowNodes is a valid switch, replace it with a component
    final SwitchComponent switchComponent = new SwitchComponent(sequenceFlows, defaultSequenceFlow);
    switchComponent.setId(Misc.getHumanReadableId("component"));
    bpmnMap.getFlowNodes().put(new QName(switchComponent.getId()), switchComponent);

    // remove nodes
    // Replace first sequenceFlow
    final TSequenceFlow firstSeqFlow = bpmnMap.createSequenceFlow(startGateway, switchComponent);

    // generate condition for the new sequenceFlow
    final TExpression tExpression = new TExpression();
    String componentCondition = "false";
    for (final TSequenceFlow seqFlow : sequenceFlows) {
      componentCondition += " or (";
      for (final TSequenceFlow otherSeqFlow : bpmnMap.getOutgoingSequenceFlows(startGateway)) {
        if (otherSeqFlow.getConditionExpression() == null) {
          componentCondition += "true)";
          break;
        }
        final ListIterator<Object> iterator = otherSeqFlow.getConditionExpression().getContent().listIterator();
        while (iterator.hasNext()) {
          final Object o = iterator.next();
          if (o instanceof String) {
            final String expr = (String) o;
            if (otherSeqFlow.equals(seqFlow)) {
              componentCondition += expr + ")";
            } else {
              componentCondition += "not(" + expr + ") and ";
            }
          }
        }
        if (otherSeqFlow.equals(seqFlow)) {
          break;
        }
      }
    }
    componentCondition = componentCondition.replace("false or ", "");
    componentCondition = componentCondition.replace(" and true", "");
    tExpression.getContent().add(componentCondition);
    firstSeqFlow.setConditionExpression(tExpression);

    // add new sequence flow first in the list
    startGateway.getOutgoings().add(0, startGateway.getOutgoings().remove(startGateway.getOutgoings().size() - 1));

    // replace last sequence flow
    bpmnMap.createSequenceFlow(switchComponent, endGateway);
    // remove sequence nodes from list
    for (final TSequenceFlow seqFlow : sequenceFlows) {
      bpmnMap.removeNode((TFlowNode) seqFlow.getTargetRef());
    }
    // check if gateways can be removed
    final Collection<TSequenceFlow> startInSeqFlows = bpmnMap.getIncomingSequenceFlows(startGateway);
    final Collection<TSequenceFlow> startOutSeqFlows = bpmnMap.getOutgoingSequenceFlows(startGateway);
    if (startInSeqFlows.size() == 1 && startOutSeqFlows.size() == 1) {
      // remove
      bpmnMap.changeSequenceFlowTarget(startInSeqFlows.iterator().next(), switchComponent);
      bpmnMap.removeNode(startGateway);
    }
    final Collection<TSequenceFlow> endInSeqFlows = bpmnMap.getIncomingSequenceFlows(endGateway);
    final Collection<TSequenceFlow> endOutSeqFlows = bpmnMap.getOutgoingSequenceFlows(endGateway);
    if (endInSeqFlows.size() == 1 && endOutSeqFlows.size() == 1) {
      // remove
      bpmnMap.changeSequenceFlowSource(endOutSeqFlows.iterator().next(), switchComponent);
      bpmnMap.removeNode(endGateway);
    }
    Misc.fastDynamicLog(SwitchTransformer.LOG, Level.INFO, "Creating switch component: %s (%s)", switchComponent.getId(), switchComponent);

    return true;
  }

  private List<TSequenceFlow> getSwitch(final BpmnMap bpmnMap, final TFlowNode startNode) {
    if (!(startNode instanceof TExclusiveGateway)) {
      return Collections.emptyList();
    }
    final Map<TGateway, List<TSequenceFlow>> sequenceFlows =
            new TreeMap<TGateway, List<TSequenceFlow>>(TFlowNodeComparator.INSTANCE);
    final List<TSequenceFlow> outgoings = bpmnMap.getOutgoingSequenceFlows(startNode);
    for (final TSequenceFlow seq : outgoings) {
      final TFlowNode target = (TFlowNode) seq.getTargetRef();
      if (target != null && !(target instanceof TExclusiveGateway)) {
        final Collection<TSequenceFlow> targetOut = bpmnMap.getOutgoingSequenceFlows(target);
        if (targetOut.size() == 1) {
          final TSequenceFlow targetSeq = targetOut.iterator().next();
          final TFlowNode nextNext = (TFlowNode) targetSeq.getTargetRef();
          if (nextNext instanceof TExclusiveGateway) {
            if (!sequenceFlows.containsKey(nextNext)) {
              sequenceFlows.put((TExclusiveGateway) nextNext, new ArrayList<TSequenceFlow>());
            }
            sequenceFlows.get(nextNext).add(seq);
          }
        }
      }
    }
    List<TSequenceFlow> result = Collections.emptyList();
    for (final List<TSequenceFlow> l : sequenceFlows.values()) {
      if (l.size() > result.size()) {
        result = l;
      }
    }
    return result;
  }

}
